/*
 *  hnv_computation_main_X11.c
 *  HeteroNumberViewer
 *
 *  Created by Naoto Morikawa on 9/15/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-10-01 version 0.1 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "hnv_main.h"
#include "hnv_load_file.h"
#include "hnv_tetrahedron.h"
#include "hnv_tetra_around_peak.h"
#include "hnv_trajectory.h"
#include "hnv_search_tree.h"
#include "hnv_cone.h"
#include "hnv_t_flow_data_strct_X11.h"	// contains #include <GLUT/glut.h>

struct parameters {
	int dim_of_lattice;
	int initial_tetra_chunk_size;
	int increment_tetra_chunk_size;
	int tetra_safe_limit;
	HNV_BOOLEAN use_extended_conj_cone;
	HNV_BOOLEAN remove_redundant_peaks;
	HNV_BOOLEAN print_out_tetra_flow;
};


//===============================================
// Private data
//===============================================

//===============================================
// Public data
//===============================================

struct parameters my_param;


//===============================================
// Private methods
//===============================================
//
// void set_up_all_lists(void);
// void initialze_resouces(void);
// HNV_TRAJ_TYPE move_trajectory_forward(struct trajectory * traj,
//					 struct hnv_cone * standard_cone,
//					 struct hnv_cone * conjugate_roof);
// void move_trajectory_backward(struct trajectory * traj, 
//				 struct hnv_cone * standard_cone, 
//				 struct hnv_cone * conjugate_roof);
// void compute_trajectory(struct tetrahedron * ft, struct hnv_cone * standard_cone, 
//			   struct hnv_cone * conjugate_roof);
//-----------------------------------------------

void set_up_all_lists(void)
{
	set_up_empty_tetra_list(&bd_tetra_list_head);
	initialize_trajectory(&open_traj_list_head);
	initialize_trajectory(&closed_traj_list_head);
	return;
}


void initialze_resouces(void)
{
	initialize_tetra_pool();
	initialize_search_node_tree();
	set_up_all_lists();
	return;
}


HNV_TRAJ_TYPE move_trajectory_forward(struct trajectory * traj,
				      struct hnv_cone * standard_cone,
				      struct hnv_cone * conjugate_roof)
{
	struct tetrahedron * ft; 
	struct tetrahedron * ft_next;
	struct tetrahedron * ft_checked;
	struct tetrahedron * ft_bd; 
	HNV_TETRA_CLASS tetra_class;
	HNV_TRAJ_TYPE trajectory_is;
	
	if (0) {
		fprintf(STDOUT, "move_trajectory_forward() : DEBUG  START of trajectory\n");
		print_trajectory(traj);
		fprintf(STDOUT, "move_trajectory_forward() : DEBUG  END of trajectory\n");
		fprintf(STDOUT, "\n");
	}
	ft = get_traj_tail(traj);	// current tetrahedron

	// Go next tetrahedron
	ft_next= assign_tetrahedron();
	compute_next_tetrahedron(ft_next, ft);
	ft_checked = check_if_processed(ft_next);
	if (ft_checked != NULL) {	// already processed
		if (get_tetra_class(ft_checked) == HNV_OUT) {
			fprintf(STDERR, "move_trajectory_forward() : ERROR come across to OUT\n");
			print_tetrahedron(ft_checked);
			exit(1);
		} else if (get_tetra_class(ft_checked) == HNV_BD) {
			if (DEBUG) {
				fprintf(STDOUT, "move_trajectory_forward() : CAUTION come across to BD\n");
				print_tetrahedron(ft_checked);
				fprintf(STDOUT, "move_trajectory_forward() :\n");
				fprintf(STDOUT, "move_trajectory_forward() : DEBUG  START of trajectory\n");
				print_trajectory(traj);
				fprintf(STDOUT, "move_trajectory_forward() : DEBUG  END of trajectory\n");
				fprintf(STDOUT, "\n");
			}
		}
		recall_tetrahedron(ft_next);
		return HNV_CLOSED;
	} else {
		compute_gradient(ft_next, standard_cone, conjugate_roof);
		tetra_class = get_tetra_class(ft_next);
		if (tetra_class  == HNV_BD) {
			ft_bd = assign_tetrahedron();
			copy_tetrahedron(ft_bd, ft_next);
			put_tetrahedron(&bd_tetra_list_head, ft_bd);
		} else if (tetra_class  == HNV_OUT) {
			recall_tetrahedron(ft_next);
			return HNV_NOT_CLOSED;	// passed by the terminal
		} else if (tetra_class != HNV_IN) {
			fprintf(STDERR, "move_trajectory_forward() : ERROR tetra type undefied (tetra_class != HNV_IN)\n");
			exit(1);
		}
		compute_direction(ft_next, ft);
		if (0) {
			fprintf(STDOUT, "move_trajectory_forward() : DEBUG START of ft_next\n");
			print_tetrahedron(ft_next);
			fprintf(STDOUT, "move_trajectory_forward() : DEBUG END of ft_next\n");
			fprintf(STDOUT, "\n");
		}
		append_to_traj_tail(traj, ft_next);
		register_in_search_tree(ft_next);
		trajectory_is = move_trajectory_forward(traj, standard_cone, conjugate_roof);
		return trajectory_is;
	}
}


void move_trajectory_backward(struct trajectory * traj, 
			      struct hnv_cone * standard_cone, 
			      struct hnv_cone * conjugate_roof)
{
	struct tetrahedron * ft; 
	struct tetrahedron * ft_previous;
	struct tetrahedron * ft_bd; 
	HNV_TETRA_CLASS tetra_class;
	
	ft = get_traj_head(traj);	// current tetrahedron
	// Go previous tetrahedron
	ft_previous = assign_tetrahedron();
	compute_previous_tetrahedron(ft_previous, ft);
	if (check_if_processed(ft_previous) != NULL) {
		recall_tetrahedron(ft_previous);
		fprintf(STDERR, "move_trajectory_backward() : ERROR trajectories intersect\n");
		exit(1);
	} else {
		compute_gradient(ft_previous, standard_cone, conjugate_roof);
		tetra_class = get_tetra_class(ft_previous);
		if (tetra_class == HNV_BD) {
			ft_bd = assign_tetrahedron();
			copy_tetrahedron(ft_bd, ft_previous);
			put_tetrahedron(&bd_tetra_list_head, ft_bd);
		} else if (tetra_class == HNV_OUT) {
			recall_tetrahedron(ft_previous);
			return;		// passed by the terminal
		} else if (tetra_class != HNV_IN) {
			fprintf(STDERR, "move_trajectory_backward() : ERROR tetra type undefied\n");
			exit(1);
		}
		compute_direction(ft_previous, ft);
		append_to_traj_head(traj, ft_previous);
		register_in_search_tree(ft_previous);
		move_trajectory_backward(traj, standard_cone, conjugate_roof);
		return;
	}
}


void compute_trajectory(struct tetrahedron * ft, struct hnv_cone * standard_cone, 
			struct hnv_cone * conjugate_roof)
{
	struct trajectory * new_traj;
	struct tetrahedron * ft_bd; 
	HNV_TETRA_CLASS tetra_class;
	HNV_TRAJ_TYPE trajectory_is;
	
	tetra_class = get_tetra_class(ft);
	if (tetra_class == HNV_BD) {
		ft_bd = assign_tetrahedron();
		copy_tetrahedron(ft_bd, ft);
		put_tetrahedron(&bd_tetra_list_head, ft_bd);
	} else if (tetra_class != HNV_IN) {
		fprintf(STDERR, "compute_trajectory() : ERROR tetra_class == HNV_OUT\n");
		exit(1);
	}
	
	set_tetra_direction(ft, HNV_DOWN);	// initial value
	new_traj = allocate_trajectory();
	if (new_traj == NULL) {
		fprintf(STDERR, 
			"compute_trajectory() : ERROR allocation failed (new_traj)\n");
		exit(1);
	}
	set_up_trajectory(new_traj, ft);
	trajectory_is = move_trajectory_forward(new_traj, standard_cone,
						conjugate_roof);
	if (trajectory_is == HNV_NOT_CLOSED) {
		move_trajectory_backward(new_traj, standard_cone,
					 conjugate_roof);
		set_traj_type(new_traj, HNV_NOT_CLOSED);
		put_trajectory(&open_traj_list_head, new_traj);
		if (0) {
			fprintf(STDOUT, "compute_trajectory() : DEBUG  START of open trajectory\n");
			print_trajectory(new_traj);
			fprintf(STDOUT, "compute_trajectory() : DEBUG  END of open trajectory\n");
			fprintf(STDOUT, "\n");
		}
	} else {
		set_traj_type(new_traj, HNV_CLOSED);
		put_trajectory(&closed_traj_list_head, new_traj);
		if (0) {
			fprintf(STDOUT, "compute_trajectory() : DEBUG  START of closed trajectory\n");
			print_trajectory(new_traj);
			fprintf(STDOUT, "compute_trajectory() : DEBUG  END of closed trajectory\n");
			fprintf(STDOUT, "\n");
		}
	}
	return;
}


//===============================================
// Public methods
//===============================================

void set_use_extended_conj_cone(HNV_BOOLEAN value)
{
	my_param.use_extended_conj_cone = value;
	return;
}


HNV_BOOLEAN get_use_extended_conj_cone(void)
{
	return my_param.use_extended_conj_cone;
}


void set_remove_redundant_peaks(HNV_BOOLEAN value)
{
	my_param.remove_redundant_peaks = value;
	return;
}


HNV_BOOLEAN get_remove_redundant_peaks(void)
{
	return my_param.remove_redundant_peaks;
}


void set_print_out_tetra_flow(HNV_BOOLEAN value)
{
	my_param.print_out_tetra_flow = value;
	return;
}


HNV_BOOLEAN get_print_out_tetra_flow(void)
{
	return my_param.print_out_tetra_flow;
}


void set_tetra_safe_limit(int value)
{
	my_param.tetra_safe_limit = value;
	return;
}


void set_up_parameters(void)
{
	my_param.initial_tetra_chunk_size = 500;
	my_param.increment_tetra_chunk_size = 200;
	my_param.tetra_safe_limit = 3000;
	my_param.use_extended_conj_cone = HNV_FALSE;
	my_param.remove_redundant_peaks = HNV_FALSE;
	my_param.print_out_tetra_flow = HNV_FALSE;
	return;
}


int get_param_safe_limit(struct parameters * pt_param)
{
	return pt_param->tetra_safe_limit;
}


int get_param_init_size(struct parameters * pt_param)
{
	return pt_param->initial_tetra_chunk_size;
}


int get_param_incre_size(struct parameters * pt_param)
{
	return pt_param->increment_tetra_chunk_size;
}


struct hnv_cone * compute_3d_case(struct hnv_cone * standard_cone)
{
	struct hnv_cone * expanded_cone;
	struct hnv_cone * conjugate_roof;
	struct tetrahedron * ft;
	struct tetrahedron * ft_work;
	struct hnv_coor_4d peak_coor;
	int p_ct, t_ct;
	HNV_TETRA_CLASS tetra_class;
	
	if (get_use_extended_conj_cone() == HNV_TRUE) {
		expanded_cone = compute_expanded_cone(standard_cone);
		if (0) {
			fprintf(STDOUT, "compute_3d_case() : DEBUG  START of expanded_cone\n");
			print_hnv_cone(expanded_cone);
			fprintf(STDOUT, "compute_3d_case() : DEBUG  END of expanded_cone\n");
			fprintf(STDOUT, "\n");
		}
		conjugate_roof = compute_conjugate_roof(expanded_cone);
		free_hnv_cone(expanded_cone);
	} else {
		conjugate_roof = compute_conjugate_roof(standard_cone);
	}
	if (0) {
		fprintf(STDOUT, "compute_3d_case() : DEBUG  START of conjugate_roof\n");
		print_hnv_cone_short(conjugate_roof);
		fprintf(STDOUT, "compute_3d_case() : DEBUG  END of conjugate_roof\n");
		fprintf(STDOUT, "\n");
	}

	initialze_resouces();

	
	for (p_ct = 0; p_ct < get_cone_peak_size(standard_cone); p_ct++) {
		if (get_peak_redundance(standard_cone, p_ct) == HNV_REDUNDANT)
			continue;
		// Compute trajectory for all 24 tetrahedrons around the peak
		get_peak_coor(&peak_coor, standard_cone, p_ct);
		tetrahedrons_around(&peak_coor);
		for (t_ct = 0; t_ct < 24; t_ct++) {
			ft = assign_tetrahedron();
			set_initial_tetra_around_peak(ft, t_ct);
			ft_work = check_if_processed(ft);
			if (ft_work == NULL) {	// NOT PROCESSED
				compute_gradient(ft, standard_cone, 
						 conjugate_roof);
				tetra_class = get_tetra_class(ft);
				if (0) {
					fprintf(STDOUT, "compute_3d_case() : DEBUG  p_ct=%d, t_ct=%d\n",
						p_ct, t_ct);
					fprintf(STDOUT, "compute_3d_case() : DEBUG START of initial tetra\n");
					print_tetrahedron(ft);
					fprintf(STDOUT, "compute_3d_case() : DEBUG END of initial tetra\n");
					fprintf(STDOUT, "\n");
				}
				if (tetra_class == HNV_OUT) {
					recall_tetrahedron(ft);
				} else {
					set_tetra_nearby_peak(ft, p_ct);
					register_in_search_tree(ft);
					compute_trajectory(ft, standard_cone, conjugate_roof);
				}
			} else {	// ft_work is already PROCESSED
				recall_tetrahedron(ft);
				set_tetra_nearby_peak(ft_work, p_ct);
			}
		}
	}
	return conjugate_roof;
}


void compute_tetrahedron_flow(const char * filename)
{	
	my_standard_cone = load_hnv_file(filename);
	if (get_remove_redundant_peaks() == HNV_TRUE)
		remove_redundant_peaks(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  START of my_standard_cone\n");
		print_hnv_cone_short(my_standard_cone);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  END of my_standard_cone\n");
		fprintf(STDOUT, "\n");
	}
	my_standard_roof = compute_roof(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  START of my_standard_roof\n");
		print_hnv_cone_short(my_standard_roof);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  END of my_standard_roof\n");
		fprintf(STDOUT, "\n");
	}
	my_conjugate_roof = compute_3d_case(my_standard_cone);
	if (0) {
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  START of my_conjugate_roof\n");
		print_hnv_cone_short(my_conjugate_roof);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  END of my_conjugate_roof\n");
		fprintf(STDOUT, "\n");
	}
	remove_redundant_peaks(my_standard_cone);
	if (0) {
		if (get_remove_redundant_peaks() != HNV_TRUE) {
			fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  START of my_standard_cone (reduced)\n");
			print_hnv_cone_short(my_standard_cone);
			fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG  END of my_standard_cone (reduced)\n");
			fprintf(STDOUT, "\n");
		}
	}
	if (0) {
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG START of closed_traj_list_head\n");
		print_traj_list(&closed_traj_list_head);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG END of closed_traj_list_head\n");
		fprintf(STDOUT, "compute_tetrahedron_flow() :\n");
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG START of open_traj_list_head\n");
		print_traj_list(&open_traj_list_head);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG END of open_traj_list_head\n");
		fprintf(STDOUT, "compute_tetrahedron_flow() :\n");
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG START of bd_tetra_list_head\n");
		print_tetra_list(&bd_tetra_list_head);
		fprintf(STDOUT, "compute_tetrahedron_flow() : DEBUG END of bd_tetra_list_head\n");
		fprintf(STDOUT, "\n");
	}
	return;
}
